# -*- coding: future_fstrings -*-
import numbers
from functools import _lru_cache_wrapper, lru_cache
from typing import *

import gridfs
from pymongo import MongoClient

from .experiment import Experiment
from .query_set import QuerySet

MAX_CACHE_SIZE = 32


class ExperimentLoader:
    """Loads artifacts related to experiments."""

    def __init__(self, mongo_uri=None, db_name="sacred"):
        client = MongoClient(mongo_uri)
        self._database = client[db_name]
        self._runs = self._database.runs
        self._grid_filesystem = gridfs.GridFS(self._database)

    def find_by_ids(self, experiment_ids: Iterable[int]) -> QuerySet:
        """
        Find experiments based on a collection of ids.

        Args:
            experiment_ids: Iterable of experiment ids.

        Returns:
            The experiments corresponding to the ids.
        """
        experiments = [self.find_by_id(experiment_id) for experiment_id in experiment_ids]

        return QuerySet(experiments)

    # The cache makes sure that retrieval of the experiments
    # is not unnecessarily done more than once.
    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_id(self, experiment_id: int) -> Experiment:
        """
        Find experiment based on its id.

        Args:
            experiment_id: The id  of the experiment.

        Returns:
            The experiment corresponding to the id.
        """
        experiment = self._find_experiment(experiment_id)

        return self._make_experiment(experiment)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_name(self, name: str) -> QuerySet:
        """
        Find experiments based on regex search against its name.

        A partial match between experiment name and regex is enough
        to find the experiment.

        Args:
            name: Regex that is matched against the experiment name.

        Returns:
            The matched experiments.
        """
        return self.find_by_key("experiment.name", name)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_config_key(self, key: str, value: Union[str, numbers.Real, tuple]) -> QuerySet:
        """
        Find experiments based on search against a configuration value.

        A partial match between configuration value and regex is enough
        to find the experiment.

        Args:
            key: Configuration key to search on.
            value: Value that is matched against the experiment's configuration.
                   Can be either a string which is  then interpreted as a regex or a number.

        Returns:
            The matched experiments.
        """
        key = f"config.{key}"
        cursor = self._search_collection(key, value)
        experiments = [self._make_experiment(experiment) for experiment in cursor]
        return QuerySet(experiments)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_by_key(self, key: str, value: Union[str, numbers.Real]) -> QuerySet:
        """
        Find experiments based on search against a value stored in the database.

        A partial match between the value and the regex is enough
        to find the experiment.

        Args:
            key: Key to search on.
            value: Value that is matched against the experiment's information.
                   Can be either a string which is  then interpreted as a regex or a number.

        Returns:
            The matched experiments.
        """
        cursor = self._search_collection(key, value)
        experiments = [self._make_experiment(experiment) for experiment in cursor]
        return QuerySet(experiments)

    @lru_cache(maxsize=MAX_CACHE_SIZE)
    def find_all(self) -> QuerySet:
        """
        Find all experiments stored in the database.

        Returns:
            All experiments.
        """
        cursor = self._runs.find()
        return QuerySet([self._make_experiment(experiment) for experiment in cursor])

    def find(self, query: dict) -> QuerySet:
        """Find experiments based on a mongo query.

        Args:
            An arbitrary mongo query.

        Returns:
            The matched experiments.
        """
        cursor = self._runs.find(query)
        experiments = [self._make_experiment(experiment) for experiment in cursor]
        return QuerySet(experiments)

    def cache_clear(self):
        """Clear all caches of all find functions.

        Useful when you want to see the updates to your database."""
        self.find_all.cache_clear()
        self.find_by_key.cache_clear()
        self.find_by_config_key.cache_clear()
        self.find_by_id.cache_clear()
        self.find_by_name.cache_clear()

    def _find_experiment(self, experiment_id: int):
        run = self._runs.find_one({"_id": experiment_id})
        if run is None:
            raise ValueError(f'Experiment with id {experiment_id} does not exist in database "{self._database.name}".')
        return run

    def _make_experiment(self, experiment) -> Experiment:
        return Experiment.from_db_object(self._database, self._grid_filesystem, experiment, loader=self)

    def _search_collection(self, key, value):
        if isinstance(value, str):
            cursor = self._runs.find({key: {"$regex": rf"{value}"}})
        elif isinstance(value, numbers.Real):
            cursor = self._runs.find({key: value})
        return cursor



# -*- coding: future_fstrings -*-
from typing import *

import pandas as pd
from easydict import EasyDict

from incense.artifact import Artifact, content_type_to_artifact_cls


class Experiment:
    def __init__(self, id_, database, grid_filesystem, data, artifact_links, loader):
        self.id = id_
        self._data = data
        self._artifacts_links = artifact_links
        self._database = database
        self._grid_filesystem = grid_filesystem
        self._loader = loader
        self._artifacts = None
        self._metrics = None

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.id}, name={self.experiment.name})"

    def __getattr__(self, item):
        """Try to relay attribute access to easy dict, to allow dotted access."""
        return getattr(self._data, item)

    @classmethod
    def from_db_object(cls, database, grid_filesystem, experiment_data: dict, loader):
        data = EasyDict(experiment_data)
        artifacts_links = experiment_data["artifacts"]
        id_ = experiment_data["_id"]
        return cls(id_, database, grid_filesystem, data, artifacts_links, loader)

    @property
    def artifacts(self) -> Dict[str, Artifact]:
        """
        The artifacts belonging to the experiment.

        Returns:
            A mapping from artifact names to artifact objects, that
            belong to the experiment.
        """
        if self._artifacts is None:
            self._artifacts = self._load_artifacts()

        return self._artifacts

    @property
    def metrics(self) -> Dict[str, pd.Series]:
        """
        The metrics belonging to the experiment.

        Returns:
            A mapping from metric names to pandas Series objects, that
            belong to the experiment.
        """
        if self._metrics is None:
            self._metrics = self._load_metrics()

        return self._metrics

    def to_dict(self) -> dict:
        """Convert the experiment to a dictionary.

        Returns:
            A dict with all data from the sacred data model.
        """
        return dict(zip(self.keys(), self.values()))

    def delete(self, confirmed: bool = False):
        """Delete run together with its artifacts and metrics.

        Args:
            confirmed: Whether to skip the confirmation prompt.
        """
        if not confirmed:
            confirmed = input(f"Are you sure you want to delete {self}? [y/N]") == "y"
        if confirmed:
            self._delete()

    def _load_artifacts(self) -> Dict[str, Artifact]:
        artifacts = {}
        for artifact_link in self._artifacts_links:
            artifact_file = self._grid_filesystem.get(artifact_link["file_id"])
            name = artifact_link["name"]

            try:
                content_type = artifact_file.content_type
                artifact_type = content_type_to_artifact_cls[content_type]
                artifacts[name] = artifact_type(name, artifact_file, content_type=content_type)
            except KeyError:
                artifact_type = Artifact
                artifacts[name] = artifact_type(name, artifact_file)

        return artifacts

    def _delete(self):
        self._delete_artifacts()
        self._delete_metrics()
        self._database.runs.delete_one({"_id": self.id})
        self._loader.cache_clear()

    def _load_metrics(self) -> Dict[str, pd.Series]:
        metrics = {}
        metric_db_entries = self._database.metrics.find({"run_id": self.id})
        for metric_db_entry in metric_db_entries:
            metrics[metric_db_entry["name"]] = pd.Series(
                data=metric_db_entry["values"],
                index=pd.Index(metric_db_entry["steps"], name="step"),
                name=metric_db_entry["name"],
            )
        return metrics

    def _delete_metrics(self):
        self._database.metrics.delete_many({"run_id": self.id})

    def _delete_artifacts(self):
        for artifact_link in self._artifacts_links:
            self._grid_filesystem.delete(artifact_link["file_id"])



import os

from setuptools import setup

import versioneer

_here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="incense",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Interactively retrieve data from sacred experiments.",
    long_description=long_description,
    author="Rüdiger Busche",
    author_email="rbusche@uos.de",
    url="https://github.com/JarnoRFB/incense",
    license="MIT",
    packages=["incense"],
    python_requires=">=3.5",
    install_requires=[
        "sacred",
        "matplotlib>=3",
        "pandas>=0.23",
        "jupyterlab>=0.35",
        "pymongo>=3.7",
        "easydict>=1.9",
        "future-fstrings==1.0.0",
    ],
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
